/* milli_httpd - pretty small HTTP server
** A combination of
** micro_httpd - really small HTTP server
** and
** mini_httpd - small HTTP server
**
** Copyright  1999,2000 by Jef Poskanzer <jef@acme.com>.
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "global_share.h" //Add by Walter, Oct-04-2004
#include <httpd.h>
#include <bcmnvram.h>
#ifdef vxworks
static void fcntl(int a, int b, int c) {}
#include <signal.h>
#include <ioLib.h>
#define socklen_t 		int
#define main			milli
#else
#include <error.h>
#include <sys/signal.h>
#endif

#define SERVER_NAME "httpd"
#define SERVER_PORT 80
#define PROTOCOL "HTTP/1.0"
#define RFC1123FMT "%a, %d %b %Y %H:%M:%S GMT"

// added by Jimmy Ma 2003.4.21
#define LOGON_IDLETIME	600
int AuthReset = 1;
// added by Jimmy Ma 2003.4.24
//Add by Terence 20030515
int MOTOROLA_Wizard = 0;
struct sockaddr_in mysockaddr;
/* A multi-family sockaddr. */
typedef union {
    struct sockaddr sa;
    struct sockaddr_in sa_in;
    } usockaddr;

/* Globals. */
static FILE *conn_fp;
static char auth_userid[AUTH_MAX];
static char auth_passwd[AUTH_MAX];
static char auth_realm[AUTH_MAX];

/* Forwards. */
static int initialize_listen_socket( usockaddr* usaP );
static int auth_check( char* dirname, char* authorization );
static void send_authenticate( char* realm );
static void send_error( int status, char* title, char* extra_header, char* text );
static void send_headers( int status, char* title, char* extra_header, char* mime_type );
static int b64_decode( const char* str, unsigned char* space, int size );
static int match( const char* pattern, const char* string );
static int match_one( const char* pattern, int patternlen, const char* string );
static void handle_request(struct sockaddr_in* newaddr);

//#define HTTPD_DEBUG

#ifdef  HTTPD_DEBUG
#define httpdprintf(str, args...) printf(str, ## args)
#else
#define httpdprintf(str, args...)
#endif

// added by Jimmy Ma 2003.4.23
static void mysigrte()
{
	
//	{
//		char msg[100];
//		char bp[20];
//    struct  tm             *rv_tm;
//    struct  timeval         curtime;
//		time_t  systime;
//		static const char *months[] = {
//		        "Jan", "Feb", "Mar", "Apr", "May", "Jun",
//		        "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
//		};
//		static const char *days[] = {
//		        "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
//		};
//
//		time(&systime);
//  	rv_tm = gmtime(&systime);
//
//  	sprintf(bp,"%s, %s %2d %2d:%02d:%02d %4d",
//            days[rv_tm->tm_wday], months[rv_tm->tm_mon], rv_tm->tm_mday, 
//            rv_tm->tm_hour, rv_tm->tm_min, rv_tm->tm_sec,
//            1900+rv_tm->tm_year);
//		
//		memset( msg, '\0', sizeof(msg) );
//		
//		sprintf( msg, "%s##%s timeout", bp, inet_ntoa(mysockaddr.sin_addr));
//		write_eventlog(2, msg);
//	}
	
	memset(&mysockaddr, 0, sizeof(mysockaddr));
	if(!AuthReset)
		AuthReset = 1;
	httpdprintf("httpd : mysigrte AuthReset is %d\n", AuthReset);
	nvram_set("login_addr","0"); //Walter add for authentication security bug
}

static int ClientCheckIn(struct sockaddr_in* addr)
{
	if(mysockaddr.sin_addr.s_addr == 0)
	{
		mysockaddr.sin_addr.s_addr = addr->sin_addr.s_addr;
		mysockaddr.sin_family = addr->sin_family;
		return 1;
	} else {
		if(mysockaddr.sin_addr.s_addr == addr->sin_addr.s_addr)
		{
			return 1;
		}
		else {
			return 0;
		}
	}
}

static int
initialize_listen_socket( usockaddr* usaP )
    {
    int listen_fd;
    int i;

    memset( usaP, 0, sizeof(usockaddr) );
    usaP->sa.sa_family = AF_INET;
    usaP->sa_in.sin_addr.s_addr = htonl( INADDR_ANY );
    usaP->sa_in.sin_port = htons( SERVER_PORT );

    listen_fd = socket( usaP->sa.sa_family, SOCK_STREAM, 0 );
    if ( listen_fd < 0 )
	{
	perror( "socket" );
	return -1;
	}
    (void) fcntl( listen_fd, F_SETFD, 1 );
    i = 1;
    if ( setsockopt( listen_fd, SOL_SOCKET, SO_REUSEADDR, (char*) &i, sizeof(i) ) < 0 )
	{
	perror( "setsockopt" );
	return -1;
	}
    if ( bind( listen_fd, &usaP->sa, sizeof(struct sockaddr_in) ) < 0 )
	{
	perror( "bind" );
	return -1;
	}
    if ( listen( listen_fd, 1024 ) < 0 )
	{
	perror( "listen" );
	return -1;
	}
    return listen_fd;
    }


static int
auth_check( char* dirname, char* authorization )
    {
    char authinfo[500];
    char* authpass;
    int l;

    /* Is this directory unprotected? */
    if ( !strlen(auth_passwd) )
	/* Yes, let the request go through. */
	return 1;

    /* Basic authorization info? */
    if ( strncmp( authorization, "Basic ", 6 ) != 0 ) {
	send_authenticate( dirname );
	return 0;
    }

    /* Decode it. */
    l = b64_decode( &(authorization[6]), authinfo, sizeof(authinfo) );
    authinfo[l] = '\0';
    /* Split into user and password. */
    authpass = strchr( authinfo, ':' );
    if ( authpass == (char*) 0 ) {
	/* No colon?  Bogus auth info. */
	send_authenticate( dirname );
	return 0;
    }
    *authpass++ = '\0';

    /* Is this the right user and password? */
    if ( strcmp( auth_userid, authinfo ) == 0 && strcmp( auth_passwd, authpass ) == 0 )
		{
			if(AuthReset)
			{
				send_authenticate( dirname );
				AuthReset = 0;
				return 0; 
			}
	
			alarm(0);
			{
				int idle_time = 0;
				
				idle_time = atoi(nvram_get("logonidletimer")) * 60;
				if(idle_time == 0)
					idle_time = 600;
					
				alarm(idle_time);
			}
			return 1;
		}

    send_authenticate( dirname );
    return 0;
    }


static void
send_authenticate( char* realm )
    {
    char header[10000];

    (void) snprintf(
	header, sizeof(header), "WWW-Authenticate: Basic realm=\"%s\"", realm );
    send_error( 401, "Unauthorized", header, "Authorization required." );
    }

static void
send_RE_authenticate( char* realm )
{
	char header[10000];
	
	(void) snprintf(
	header, sizeof(header), "WWW-Authenticate: Basic realm=\"%s\"", realm );
	send_error( 403.17, "Timeout", header, "Authorization required." );
}

static void
send_error( int status, char* title, char* extra_header, char* text )
    {
    send_headers( status, title, extra_header, "text/html" );
    (void) fprintf( conn_fp, "<HTML><HEAD><TITLE>%d %s</TITLE></HEAD>\n<BODY BGCOLOR=\"#cc9999\"><H4>%d %s</H4>\n", status, title, status, title );
    (void) fprintf( conn_fp, "%s\n", text );
    (void) fprintf( conn_fp, "</BODY></HTML>\n" );
    (void) fflush( conn_fp );
    }


static void
send_headers( int status, char* title, char* extra_header, char* mime_type )
    {
    time_t now;
    char timebuf[100];

    (void) fprintf( conn_fp, "%s %d %s\r\n", PROTOCOL, status, title );
    (void) fprintf( conn_fp, "Server: %s\r\n", SERVER_NAME );
    now = time( (time_t*) 0 );
    (void) strftime( timebuf, sizeof(timebuf), RFC1123FMT, gmtime( &now ) );
    (void) fprintf( conn_fp, "Date: %s\r\n", timebuf );
    if ( extra_header != (char*) 0 )
	(void) fprintf( conn_fp, "%s\r\n", extra_header );
    if ( mime_type != (char*) 0 )
	(void) fprintf( conn_fp, "Content-Type: %s\r\n", mime_type );
    (void) fprintf( conn_fp, "Connection: close\r\n" );
    (void) fprintf( conn_fp, "\r\n" );
    }


/* Base-64 decoding.  This represents binary data as printable ASCII
** characters.  Three 8-bit binary bytes are turned into four 6-bit
** values, like so:
**
**   [11111111]  [22222222]  [33333333]
**
**   [111111] [112222] [222233] [333333]
**
** Then the 6-bit values are represented using the characters "A-Za-z0-9+/".
*/

static int b64_decode_table[256] = {
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* 00-0F */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* 10-1F */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,62,-1,-1,-1,63,  /* 20-2F */
    52,53,54,55,56,57,58,59,60,61,-1,-1,-1,-1,-1,-1,  /* 30-3F */
    -1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,  /* 40-4F */
    15,16,17,18,19,20,21,22,23,24,25,-1,-1,-1,-1,-1,  /* 50-5F */
    -1,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,  /* 60-6F */
    41,42,43,44,45,46,47,48,49,50,51,-1,-1,-1,-1,-1,  /* 70-7F */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* 80-8F */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* 90-9F */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* A0-AF */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* B0-BF */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* C0-CF */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* D0-DF */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,  /* E0-EF */
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1   /* F0-FF */
    };

/* Do base-64 decoding on a string.  Ignore any non-base64 bytes.
** Return the actual number of bytes generated.  The decoded size will
** be at most 3/4 the size of the encoded, and may be smaller if there
** are padding characters (blanks, newlines).
*/
static int
b64_decode( const char* str, unsigned char* space, int size )
    {
    const char* cp;
    int space_idx, phase;
    int d, prev_d=0;
    unsigned char c;

    space_idx = 0;
    phase = 0;
    for ( cp = str; *cp != '\0'; ++cp )
	{
	d = b64_decode_table[(int)*cp];
	if ( d != -1 )
	    {
	    switch ( phase )
		{
		case 0:
		++phase;
		break;
		case 1:
		c = ( ( prev_d << 2 ) | ( ( d & 0x30 ) >> 4 ) );
		if ( space_idx < size )
		    space[space_idx++] = c;
		++phase;
		break;
		case 2:
		c = ( ( ( prev_d & 0xf ) << 4 ) | ( ( d & 0x3c ) >> 2 ) );
		if ( space_idx < size )
		    space[space_idx++] = c;
		++phase;
		break;
		case 3:
		c = ( ( ( prev_d & 0x03 ) << 6 ) | d );
		if ( space_idx < size )
		    space[space_idx++] = c;
		phase = 0;
		break;
		}
	    prev_d = d;
	    }
	}
    return space_idx;
    }


/* Simple shell-style filename matcher.  Only does ? * and **, and multiple
** patterns separated by |.  Returns 1 or 0.
*/
int
match( const char* pattern, const char* string )
    {
    const char* or;

    for (;;)
	{
	or = strchr( pattern, '|' );
	if ( or == (char*) 0 )
	    return match_one( pattern, strlen( pattern ), string );
	if ( match_one( pattern, or - pattern, string ) )
	    return 1;
	pattern = or + 1;
	}
    }


static int
match_one( const char* pattern, int patternlen, const char* string )
    {
    const char* p;

    for ( p = pattern; p - pattern < patternlen; ++p, ++string )
	{
	if ( *p == '?' && *string != '\0' )
	    continue;
	if ( *p == '*' )
	    {
	    int i, pl;
	    ++p;
	    if ( *p == '*' )
		{
		/* Double-wildcard matches anything. */
		++p;
		i = strlen( string );
		}
	    else
		/* Single-wildcard matches anything but slash. */
		i = strcspn( string, "/" );
	    pl = patternlen - ( p - pattern );
	    for ( ; i >= 0; --i )
		if ( match_one( p, pl, &(string[i]) ) )
		    return 1;
	    return 0;
	    }
	if ( *p != *string )
	    return 0;
	}
    if ( *string == '\0' )
	return 1;
    return 0;
    }


void
do_file(char *path, FILE *stream)
{
	FILE *fp;
	int c;

	if (!(fp = fopen(path, "r")))
		return;
	while ((c = getc(fp)) != EOF)
		fputc(c, stream);
	fclose(fp);
}

void redirect(char *to_url)
{
	char header[200];
	httpdprintf("httpd : dirname(auth_realm) is %s\n", auth_realm);
	sprintf(header, "Location: %s", to_url);
	send_headers( 302, "Redirect", header, "text/xml" );
	fflush(conn_fp);
	fclose(conn_fp);
}
						
static void
handle_request(struct sockaddr_in* newaddr)
{
    char line[10000], method[10000], path[10000], protocol[10000], authorization[1000], boundary[1000];
    char *cp;
    char *file;
    int len;
    struct mime_handler *handler;
    int cl = 0, flags;
		char buff[50];
		
		MOTOROLA_Wizard = 0;
		
    /* Initialize the request variables. */
    strcpy(authorization, "");
    strcpy(boundary, "");

    /* Parse the first line of the request. */
    if ( fgets( line, sizeof(line), conn_fp ) == (char*) 0 ) {
	send_error( 400, "Bad Request", (char*) 0, "No request found." );
	return;
    }
    if ( sscanf( line, "%[^ ] %[^ ] %[^ ]", method, path, protocol ) != 3 ) {
	send_error( 400, "Bad Request", (char*) 0, "Can't parse request." );
	return;
    }
		
    /* Parse the rest of the request headers. */
    
    while ( fgets( line, sizeof(line), conn_fp ) != (char*) 0 )
	{
		char *pUA = 0;
		/* Add by Terence 20030515 for Motorola Setup Wizard */
		if((pUA = strstr(line, "User-Agent")) != 0)
		{
			if(strstr(pUA, "Motorola/1.0 HomeNet/1.0") != 0)
			//if(strstr(pUA, "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.0; .NET CLR 1.0.3705)") != 0)
			{
				//printf("httpd : This is IE Browser\n");
				MOTOROLA_Wizard = 1;
			}
		}
		/* end */
		
		if ( strcmp( line, "\n" ) == 0 || strcmp( line, "\r\n" ) == 0 )
		    break;
		else if ( strncasecmp( line, "Authorization:", 14 ) == 0 )
		    {
		    cp = &line[14];
		    cp += strspn( cp, " \t" );
		    strncpy( authorization, cp, sizeof(authorization) );
		    }
		else if (strncasecmp( line, "Content-Length:", 15 ) == 0) {
			cp = &line[15];
			cp += strspn( cp, " \t" );
			cl = strtoul( cp, NULL, 0 );
		}
		else if ((cp = strstr( line, "boundary=" ))) {
		    strncpy( boundary, &cp[9], sizeof(boundary) );
		    if ((cp = strchr( boundary, '\r' ))) *cp = '\0';
		    if ((cp = strchr( boundary, '\n' ))) *cp = '\0';
		}
	}
    if ( strcasecmp( method, "get" ) != 0 && strcasecmp(method, "post") != 0 ) {
	send_error( 501, "Not Implemented", (char*) 0, "That method is not implemented." );
	return;
    }
    if ( path[0] != '/' ) {
	send_error( 400, "Bad Request", (char*) 0, "Bad filename." );
	return;
    }
    file = &(path[1]);
    len = strlen( file );
    if ( file[0] == '/' || strcmp( file, ".." ) == 0 || strncmp( file, "../", 3 ) == 0 || strstr( file, "/../" ) != (char*) 0 || strcmp( &(file[len-3]), "/.." ) == 0 ) {
	send_error( 400, "Bad Request", (char*) 0, "Illegal filename." );
	return;
    }
    if ( file[0] == '\0' || file[len-1] == '/' )
	(void) snprintf( &file[len], sizeof(path) - len - 1, "index.asp" );
    path[sizeof(path) - 1] = '\0';

#if defined(vxworks)
    if (match("*x_internetgatewaydevice.xml*", file)) {
	    extern struct in_addr *osl_ifaddr(const char *ifname, struct in_addr *inaddr);
	    struct in_addr inaddr;
	    const unsigned char *bytep;
	    char header[200];

	    if ( !osl_ifaddr("br0", &inaddr) ) {
		strcpy(header, "");
	    } else {
		bytep = (const unsigned char *) &(inaddr.s_addr);
		sprintf(header, "Location: http://%d.%d.%d.%d:5431/x_internetgatewaydevice.xml", bytep[0], bytep[1], bytep[2], bytep[3]);
	    }
	    send_headers( 302, "Redirect", header, "text/xml" );
	    return;
    }
#endif
    for (handler = &mime_handlers[0]; handler->pattern; handler++) {
	    if (match(handler->pattern, file)) {
	    	int allow_file = 0;
/*** Mark by Terence for Motorola. Use WEB UI Authentication method	***/	    	
		    //if (handler->auth) {
			  //  handler->auth(auth_userid, auth_passwd, auth_realm);
				httpdprintf("httpd : file name is  : %s\n", file);
				
		  	httpdprintf("httpd : AuthReset => %d\n", AuthReset);
				
				
				if((MOTOROLA_Wizard != 1) && (strstr(file, ".html") == 0))
				{
					//printf("httpd: Other browser General in...\n");
				}		
				else if((MOTOROLA_Wizard == 1) && (strstr(file, ".html") == 0))
				{
					//printf("httpd: Motorola General in...\n");
				}								
				else if((MOTOROLA_Wizard != 1) && (strstr(file, ".html") != 0))
				{
					//printf("httpd: Other browser want to access Motorola Wizard...\n");
					send_error( 404, "Not Found", (char*) 0, "File not found." );
					return;
				}
				else if((MOTOROLA_Wizard == 1) && (strstr(file, ".html") != 0))
				{
					//printf("httpd: Motorola Wizard in...\n");
					//allow_file = 1;		
				}
				
				if((!strcmp(file, "index.asp")) || (!strcmp(file, "sysconf.cgi")))
				{							
					//nvram_set("login_addr", inet_ntoa(newaddr->sin_addr)); //Walter comment							
					memset( login_addr_global, 0, sizeof(login_addr_global));
					strcpy( login_addr_global,inet_ntoa(newaddr->sin_addr));									
					httpdprintf("httpd : login_addr is %s\n", login_addr_global);
				}	
				
				if((strstr(file, ".gif") != NULL) || (strstr(file, ".css") != NULL) || (!strcmp(file, "redirect.asp")) || (!strcmp(file, "index.asp")) || (!strcmp(file, "sysconf.cgi")) )
					allow_file = 1;
					
				if(!strcmp(file, "index.asp"))
				{
//					{
//						char msg[100];
//						char bp[20];
//			      struct  tm             *rv_tm;
//			      struct  timeval         curtime;
//						time_t  systime;
//						static const char *months[] = {
//						        "Jan", "Feb", "Mar", "Apr", "May", "Jun",
//						        "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
//						};
//						static const char *days[] = {
//						        "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
//						};
//			
//						time(&systime);
//			    	rv_tm = gmtime(&systime);
//			
//			    	sprintf(bp,"%s, %s %2d %2d:%02d:%02d %4d",
//			              days[rv_tm->tm_wday], months[rv_tm->tm_mon], rv_tm->tm_mday, 
//			              rv_tm->tm_hour, rv_tm->tm_min, rv_tm->tm_sec,
//			              1900+rv_tm->tm_year);
//						
//						memset( msg, '\0', sizeof(msg) );
//						
//						sprintf( msg, "%s##%s access", bp, nvram_get("login_addr"));
//						write_eventlog(2, msg);
//					}
				}
				
			if(allow_file != 1) // if not gif or css or redirect.asp or index.asp => do auth
			{
				httpdprintf("httpd : allow_file != 1\n");								        				         	
			       	//fprintf(stderr," inet_ntoa(newaddr->sin_addr)= %s \n",inet_ntoa(newaddr->sin_addr));
			       	//fprintf(stderr," login_addr = %s \n",nvram_get("login_addr"));
			       	fprintf(stderr,"AuthReset: %d \n", AuthReset);				
			         	
			         /*Walter rewirte begin 2004-Oct-04 */	
				if( strcmp(nvram_get("login_addr"),inet_ntoa(newaddr->sin_addr))) // check login_addr and income ip here
				 {
			  	
		    		   char header[200];
	  	  	  	   httpdprintf("httpd : Deny Client in\n");		
			  	   sprintf(header, "Location: redirect.asp");
				   send_headers( 302, "Redirect", header, "text/xml" );    	
	       		           //sprintf(buff, "The existing client:%s", inet_ntoa(mysockaddr.sin_addr.s_addr));	
			           send_error(403.9, "Only one login allowed", NULL, buff);
			           return;
		  		 }
			 	 else
			  	{	
			  	 if (AuthReset)
		                    {						
				 	char header[200];
					fprintf(stderr,"READY TO REDIRECT\n");
					//httpdprintf("httpd : Allow Client in\n");
			  		httpdprintf("httpd : dirname(auth_realm) is %s\n", auth_realm);
			  		//send_RE_authenticate( auth_realm );
			  		sprintf(header, "Location: redirect.asp");
				    	send_headers( 302, "Redirect", header, "text/xml" );
					AuthReset = 1;
					httpdprintf("httpd : handle_request AuthReset is %d\n", AuthReset);						
					return 0; 
				     }
				  else
				     { //for debug
					//fprintf(stderr,"pass! Start load page.. \n");	
				     }
				    /*Walter rewirte end 2004-Oct-04 */	
				    alarm(0);
				     {
						int idle_time = 0;
						
						idle_time = atoi(nvram_get("logonidletimer")) * 60;
						if(idle_time == 0)
							idle_time = 600;
							
						httpdprintf("httpd : idle time is %d\n", idle_time);
						alarm(idle_time);
				     }
				}

		  }
		    if (strcasecmp(method, "post") == 0 && !handler->input) {
			    send_error(501, "Not Implemented", NULL, "That method is not implemented.");
			    return;
		    }			    
		    if (handler->input) {
			    handler->input(file, conn_fp, cl, boundary);
#if defined(linux)
			    if ((flags = fcntl(fileno(conn_fp), F_GETFL)) != -1 &&
				fcntl(fileno(conn_fp), F_SETFL, flags | O_NONBLOCK) != -1) {
				    /* Read up to two more characters */
				    if (fgetc(conn_fp) != EOF)
					    (void) fgetc(conn_fp);
				    fcntl(fileno(conn_fp), F_SETFL, flags);
			    }
#elif defined(vxworks)
			    flags = 1;
			    if (ioctl(fileno(conn_fp), FIONBIO, (int) &flags) != -1) {
				    /* Read up to two more characters */
				    if (fgetc(conn_fp) != EOF)
					    (void) fgetc(conn_fp);
				    flags = 0;
				    ioctl(fileno(conn_fp), FIONBIO, (int) &flags);
			    }
#endif				
		    }
		    send_headers( 200, "Ok", handler->extra_header, handler->mime_type );
		    if (handler->output)
			    handler->output(file, conn_fp);
		    break;
	    }
    }

    if (!handler->pattern) 
	    send_error( 404, "Not Found", (char*) 0, "File not found." );
}


int main(int argc, char **argv)
{
	usockaddr usa;
	int listen_fd;
	int conn_fd;
	socklen_t sz = sizeof(usa);

/*Walter add, Jun. 16, 2004 */
/*Walter move from end, July. 09, 2004 */
	if(!strcmp("WA840GP",nvram_safe_get("ModelId")))
	{
	
               system("/bin/ln -sf /www/log_in_bg_fail_GP.gif /tmp/log_in_bg_fail.gif");
  	       system("/bin/ln -sf /www/log_in_bg_GP.gif      /tmp/log_in_bg.gif");
               system("/bin/ln -sf /www/main_bottom_center-all_GP.gif  /tmp/main_bottom_center-all.gif"); 
               system("/bin/ln -sf /www/main_left-frame1_logo_GP.gif  /tmp/main_left-frame1_logo.gif");	
         
        }
        else 
        {
               system("/bin/ln -sf /www/log_in_bg_fail_G.gif /tmp/log_in_bg_fail.gif");
  	       system("/bin/ln -sf /www/log_in_bg_G.gif      /tmp/log_in_bg.gif");
               system("/bin/ln -sf /www/main_bottom_center-all_G.gif  /tmp/main_bottom_center-all.gif"); 
               system("/bin/ln -sf /www/main_left-frame1_logo_G.gif /tmp/main_left-frame1_logo.gif");         	        	
        }
        /*end of Walter add */
	
	
	//AuthReset = 1;
	httpdprintf("httpd : Start AuthReset is %d %p\n", AuthReset, &AuthReset);
	/* Ignore broken pipes */
	signal(SIGPIPE, SIG_IGN);
	signal(SIGALRM, mysigrte);	//added by Jimmy Ma 2003.4.23

	/* Initialize listen socket */
	if ((listen_fd = initialize_listen_socket(&usa)) < 0) {
		fprintf(stderr, "can't bind to any address\n" );
		exit(errno);
	}

#if !defined(DEBUG) && !defined(vxworks)
	{
	FILE *pid_fp;
	/* Daemonize and log PID */
	if (daemon(1, 1) == -1) {
		perror("daemon");
		exit(errno);
	}
	if (!(pid_fp = fopen("/var/run/httpd.pid", "w"))) {
		perror("/var/run/httpd.pid");
		return errno;
	}
	fprintf(pid_fp, "%d", getpid());
	fclose(pid_fp);
	}
#endif

	memset(&mysockaddr, 0, sizeof(mysockaddr));

        
	/* Loop forever handling requests */
	for (;;) {
		if ((conn_fd = accept(listen_fd, &usa.sa, &sz)) < 0) {
			perror("accept");
			return errno;
		}
		if (!(conn_fp = fdopen(conn_fd, "r+"))) {
			perror("fdopen");
			return errno;
		}
		handle_request(&usa.sa);
		fflush(conn_fp);
		fclose(conn_fp);
		close(conn_fd);
	}

	shutdown(listen_fd, 2);
	close(listen_fd);

	return 0;
}
