/*
 * Copyright 2001-2005, Broadcom Corporation
 * All Rights Reserved.
 * 
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id$
 *
*/
 
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <signal.h>
#include <string.h>
#include <sys/klog.h>
#include <sys/types.h>
#include <sys/mount.h>
#include <sys/reboot.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <sys/wait.h>
#include <dirent.h>

#include <epivers.h>
#include <bcmnvram.h>
#include <mtd.h>
#include <shutils.h>
#include <rc.h>
#include <netconf.h>
#include <nvparse.h>
#include <bcmdevs.h>

#include <sys/socket.h>
#include <sys/sysinfo.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if_arp.h>

#include <wlioctl.h>
#include <wlutils.h>
#include <sys/ioctl.h>

#include "./gemtek_added/ExecutingTable.h"		/* Added by PeterChen */
#include "./gemtek_added/ConvParas.h"		/* Added by PeterChen */
#include "./gemtek_added/gemtek.h"
#include "./gemtek_added/Peter_Added_001.h"
	
char	wan_done_flag = 0;	/* Added by PeterChen, Gemtek 2003-0702 PM 08:37 */
void main_loop_stop_1(void);
void main_loop_stop_2(void);
void main_loop_start_1(void);
void main_loop_start_2(void);
static void restore_defaults(void);
static void sysinit ( void );
static void rc_signal(int sig);

extern struct nvram_tuple	router_defaults[];

static void
restore_defaults(void)
{
#if 0
	struct nvram_tuple generic[] = {
		{ "lan_ifname", "br0", 0 },
		{ "lan_ifnames", "eth0 eth2 eth3 eth4", 0 },
		{ "wan_ifname", "eth1", 0 },
		{ "wan_ifnames", "eth1", 0 },
		{ 0, 0, 0 }
	};
	struct nvram_tuple vlan[] = {
		{ "lan_ifname", "br0", 0 },
		{ "lan_ifnames", "vlan0 eth1 eth2 eth3", 0 },
		{ "wan_ifname", "vlan1", 0 },
		{ "wan_ifnames", "vlan1", 0 },
		{ 0, 0, 0 }
	};
	struct nvram_tuple *linux_overrides;
	struct nvram_tuple *t, *u;
	int restore_defaults, i;
	uint boardflags;

	/* Restore defaults if told to or OS has changed */
	restore_defaults = !nvram_match("restore_defaults", "0") || nvram_invmatch("os_name", "linux");
	if (restore_defaults)
		cprintf("Restoring defaults...");

	/* Delete dynamically generated variables */
	if (restore_defaults) {
		char tmp[100], prefix[] = "wlXXXXXXXXXX_";
		for (i = 0; i < MAX_NVPARSE; i++) {
			del_filter_client(i);
			del_forward_port(i);
			del_autofw_port(i);
			snprintf(prefix, sizeof(prefix), "wl%d_", i);
			for (t = router_defaults; t->name; t ++) {
				if (!strncmp(t->name, "wl_", 3))
					nvram_unset(strcat_r(prefix, &t->name[3], tmp));
			}
			snprintf(prefix, sizeof(prefix), "wan%d_", i);
			for (t = router_defaults; t->name; t ++) {
				if (!strncmp(t->name, "wan_", 4))
					nvram_unset(strcat_r(prefix, &t->name[4], tmp));
			}
		}
	}

	/* Choose default lan/wan i/f list. */
	boardflags = strtoul(nvram_safe_get("boardflags"), NULL, 0);
	if (boardflags & BFL_ENETVLAN)
		linux_overrides = vlan;
	else
		linux_overrides = generic;

	/* Restore defaults */
	for (t = router_defaults; t->name; t++) {
		if (restore_defaults || !nvram_get(t->name)) {
			for (u = linux_overrides; u && u->name; u++) {
				if (!strcmp(t->name, u->name)) {
					nvram_set(u->name, u->value);
					break;
				}
			}
			if (!u || !u->name)
				nvram_set(t->name, t->value);
		}
	}

	/* Always set OS defaults */
	nvram_set("os_name", "linux");
	nvram_set("os_version", EPI_ROUTER_VERSION_STR);

	/* Commit values */
	if (restore_defaults) {
		nvram_commit();
		cprintf("done\n");
	}
#endif	
}

static void
set_wan0_vars(void)
{
#if 0	
	int unit;
	char tmp[100], prefix[] = "wanXXXXXXXXXX_";
	
	/* check if there are any connections configured */
	for (unit = 0; unit < MAX_NVPARSE; unit ++) {
		snprintf(prefix, sizeof(prefix), "wan%d_", unit);
		if (nvram_get(strcat_r(prefix, "unit", tmp)))
			break;
	}
	/* automatically configure wan0_ if no connections found */
	if (unit >= MAX_NVPARSE) {
		struct nvram_tuple *t;
		char *v;

		/* Write through to wan0_ variable set */
		snprintf(prefix, sizeof(prefix), "wan%d_", 0);
		for (t = router_defaults; t->name; t ++) {
			if (!strncmp(t->name, "wan_", 4)) {
				if (nvram_get(strcat_r(prefix, &t->name[4], tmp)))
					continue;
				v = nvram_get(t->name);
				nvram_set(tmp, v ? v : t->value);
			}
		}
		nvram_set(strcat_r(prefix, "unit", tmp), "0");
		nvram_set(strcat_r(prefix, "desc", tmp), "Default Connection");
		nvram_set(strcat_r(prefix, "primary", tmp), "1");
	}
#endif	
}

static int		noconsole = 0;

static void	
sysinit ( void )
{
	char		buf [ PATH_MAX ];
	struct utsname	name;
	struct stat	tmp_stat;
	time_t		tm = 0;

	/* /proc */
	mount ( "proc" , "/proc" , "proc" , MS_MGC_VAL , NULL );

	/* /tmp */
	mount ( "ramfs" , "/tmp" , "ramfs" , MS_MGC_VAL, NULL );

	/* /var */
	mkdir ( "/tmp/var" , 0777 );
	mkdir ( "/var/lock" , 0777 );
	mkdir("/var/log", 0777);
	mkdir("/var/run", 0777);
	mkdir("/var/tmp", 0777);

	/* Setup console */
	if ( console_init () )
		noconsole = 1;
	klogctl ( 8 , NULL , atoi ( nvram_safe_get ( "console_loglevel" ) ) );

	/* Get CPU info */
	check_brcm_cpu_type();
	
	convert_wireless_params();	
	
	/* Modules */
	uname ( &name );
	snprintf ( buf , sizeof ( buf ) , "/lib/modules/%s" , name.release );
	if (stat("/proc/modules", &tmp_stat) == 0 &&
	    stat(buf, &tmp_stat) == 0) {
		char module[80], *modules, *next;
		//modules = nvram_get("kernel_mods") ? : "et il wl";
		/*Yao Chung Added*/
		char *modulestr;
		if ((atoi(nvram_safe_get("WlEnable")) == 0 ) && (atoi(nvram_safe_get("Wl5GEnable")) == 0 ))
			modulestr = "kmemory et";
		else if (atoi(nvram_safe_get("Wl5GEnable")) == 0 )
			modulestr = "kmemory et wl";
		else if	(atoi(nvram_safe_get("WlEnable")) == 0 )
		{
			//Remove by Terence, 20030507
			modulestr = "kmemory et";
			//modulestr = "kmemory et tulip tulip_1";
		}
		else
		{
			//Remove by Terence, 20030507
			modulestr = "kmemory et wl";
			//modulestr = "kmemory et wl tulip tulip_1";
		}
		modules = modulestr;
		Peter_Added_001_H_Extra_SRCs_001

		/*Yao Chung Added*/	
		/*modules = nvram_get("kernel_mods") ? : "kmemory et il wl tulip tulip_1";
		*/
		foreach(module, modules, next)
			eval("insmod", module);
//		Peter_Added_001_H_Extra_SRCs_003
	}

	/* Set a sane date */
	stime ( &tm );

	dprintf("done\n");
}

/* States */
enum {
	RESTART,
	STOP,
	START,
	TIMER,
	IDLE,
};
static int state = START;
static int signalled = -1;

/* Signal handling */
static void
rc_signal(int sig)
{
	if (state == IDLE) {	
		if (sig == SIGHUP) {
			dprintf("signalling RESTART\n");
			signalled = RESTART;
		}
		else if (sig == SIGUSR2) {
			dprintf("signalling START\n");
			signalled = START;
		}
		else if (sig == SIGINT) {
			dprintf("signalling STOP\n");
			signalled = STOP;
		}
		else if (sig == SIGALRM) {
			dprintf("signalling TIMER\n");
			signalled = TIMER;
		}
	}
}

/* Timer procedure */
int
do_timer(void)
{
	int interval = atoi(nvram_safe_get("timer_interval"));
	time_t now;
	struct tm gm, local;
	struct timezone tz;

	dprintf("%d\n", interval);

	if (interval == 0)
		return 0;

	/* Report stats */
	if (nvram_invmatch("stats_server", "")) {
		char *stats_argv[] = { "stats", nvram_get("stats_server"), NULL };
		_eval(stats_argv, NULL, 5, NULL);
	}

	/* Sync time */
	//start_ntpc();

	/* Update kernel timezone */
	setenv("TZ", nvram_safe_get("time_zone"), 1);
	time(&now);
	gmtime_r(&now, &gm);
	localtime_r(&now, &local);
	tz.tz_minuteswest = (mktime(&gm) - mktime(&local)) / 60;
	settimeofday(NULL, &tz);

	alarm(interval);
	return 0;
}

/* Main loop */
static void	
main_loop ( void )
{
	sigset_t	sigset;
	pid_t	shell_pid = 0;
	uint boardflags;
	
	/* Basic initialization */
	sysinit ();
	
	//convert_wireless_params();	

	/* Setup signal handlers */
	signal_init ();
	signal ( SIGHUP , rc_signal );
	signal ( SIGUSR2 , rc_signal );
	signal ( SIGINT , rc_signal );
	signal ( SIGALRM , rc_signal );
	sigemptyset ( &sigset );

	/* Give user a chance to run a shell before bringing up the rest of the system */
	//if (!noconsole)
	//	run_shell(1, 0);
	
	/* Convert deprecated variables */
	//convert_deprecated();	
	
	/* Restore defaults if necessary */
	//restore_defaults();	
	
	/* Marked by YaoChung*/
	gemtek_restore_defaults();

	// start_upgrade_pmon();
	/*Marked by YaoChung*/

	/* Setup wan0 variables if necessary */
//	set_wan0_vars();

	/* Add loopback */
	config_loopback();
	
//	/* Add vlan */
//	boardflags = strtoul(nvram_safe_get("boardflags"), NULL, 0);
//	if (boardflags & BFL_ENETVLAN)
//		config_vlan();

	/* Loop forever */
	for ( ; ; ) {
		switch ( state ) {
		case	RESTART:
			dprintf ( "RESTART\n" );
			/* Fall through */
		case	STOP:
			dprintf ( "STOP\n" );
			//stop_services();
			main_loop_stop_1();
			//stop_wan ();
			main_loop_stop_2();
			sleep ( 2 );
			//stop_lan();
			if ( state == STOP ) {
				state = IDLE;
				break;
			}
			/* Fall through */
		case START:
			//dprintf("START\n");
			main_loop_start_1();
			start_lan ();
			//start_services();
			/* start_wan and start_nas move to main_loop_start_2() */
			//start_wan();
			//start_nas("wan");			
			main_loop_start_2();
		case TIMER:
			dprintf("TIMER\n");
			do_timer();
			/* Fall through */
		case IDLE:
			dprintf("IDLE\n");
			state = IDLE;
			/* Wait for user input or state change */
			while (signalled == -1) {
				if (!noconsole && (!shell_pid || kill(shell_pid, 0) != 0))
					shell_pid = run_shell(0, 1);
				else
					sigsuspend(&sigset);
			}
			state = signalled;
			signalled = -1;
			break;
		default:
			dprintf("UNKNOWN\n");
			return;
		}
	}

	/* Give user a chance to run a shell before bringing up the rest of the system */
	if (!noconsole)
		run_shell (1 , 0);
}

int	
main ( int argc , char **argv )
{
	char	*base = strrchr ( argv[0] , '/' );

	base = base ? base + 1 : argv[0];
//	nvram_set ( "wan_lease" , "*******" );

	/* init */
	if ( strstr ( base , "init" ) ) {
		main_loop ();
		//return 0;
	} 
	
	/* Set TZ for all rc programs */
	//setenv("TZ", nvram_safe_get("time_zone"), 1);
	
//	/* ppp */
//	if ( strstr ( base , "ip-up" ) )	
//		return	ipup_main ( argc , argv );
//	else if ( strstr ( base , "ip-down" ) )
//		return	ipdown_main ( argc , argv );
//		
//	
//	else 

	/* udhcpc [ deconfig bound renew ] */
	if ( strstr ( base , "udhcpc" ) )	
		return	udhcpc_main ( argc , argv );
		
	/* stats [ url ] */		
	else if ( strstr ( base , "stats" ) )	
		return	http_stats ( argv[1] ? : nvram_safe_get ( "stats_server" ) );

#ifdef MTD
	/* erase [device] */
	else if ( strstr ( base , "erase" ) ) {	
		if ( argv[1] )
			return	mtd_erase ( argv[1] );
		else {
			fprintf ( stderr , "usage: erase [device]\n" );
			return	EINVAL;
		}
	}
	
	/* write [path] [device] */
	else if ( strstr ( base , "write" ) ) {	
		if ( argc >= 3 )
			return	mtd_write ( argv[1] , argv[2] );
		else {
			fprintf ( stderr , "usage: write [path] [device]\n" );
			return	EINVAL;
		}
	}
#endif
	/* hotplug [event] */
	else if ( strstr ( base , "hotplug" ) ) {	
		if ( argc >= 2 ) {
			if ( !strcmp ( argv[1] , "net" ) )
				return	hotplug_net ();
		} else {
			fprintf ( stderr , "usage: hotplug [event]\n" );
			return	EINVAL;
		}
	} 
	
	/* rc [stop|start|restart ] */
	else if ( strstr ( base , "rc" ) ) {	
		if ( argv[1] ) {
			if ( strncmp ( argv[1] , "start" , 5 ) == 0 )
				return	kill ( 1 , SIGUSR2 );
			else if (strncmp(argv[1], "stop", 4) == 0)
				return	kill ( 1 , SIGINT );
			else if (strncmp(argv[1], "restart", 7) == 0)
				return	kill ( 1 , SIGHUP );
		} else {
			fprintf ( stderr , "usage: rc [start|stop|restart]\n" );
			return	EINVAL;
		}
	} 
//	else if ( strstr ( base , "start_wan" ) )	/* start & stop wan connection */
//		start_wan();
//	else if ( strstr ( base , "stop_wan" ) )
//		stop_wan();
//	else if ( strstr ( base , "GetPPTPAddr" ) )/*get pptp address*/
//		GetPPTPAddr();
	else if ( strstr ( base , "assoclistd" ) ) {		/* Modify by Terence from Dainel */
		start_wl_assoclist ();
	}
	else if ( strstr ( base , "sitesurvey" ) ) {
		wlconfig ( nvram_safe_get("wlif") );
	}		

	return EINVAL;
}

void rc_system ( const char *string )
{
	char	i = 10;

	if ( string == NULL ) { return; }

	while ( i > 0 ) {
		i--;
		if ( system ( string ) != -1 ) {
			if ( i < 10 ) { /* printf ( "Re-executing utility '%s' ... Success\n" , string ); */ }
			return;
		} else {
			/* printf ( "Executing utility '%s' ... Failure\n" , string ); */
			sleep ( 10 - i );	/* Added by PeterChen , Gemtek , 2003-0113-16:51 */
		}
	}

	printf ( "system(%s) fails to execute with errno=%d\n" , string , errno );
}

void main_loop_stop_1(void)
{
	netconf_reset_fw ();
	//stop_firewall ();
	stop_nas();
}

void main_loop_stop_2(void)
{
		OpenFile ();	/* Added by PeterChen */
		WriteFile ( "/usr/bin/killall udhcpc" );
		//WriteFile ( "/usr/bin/killall easyconf" );
		WriteFile ( "/usr/bin/killall ResetToDefault" );
		//WriteFile ( "/usr/bin/killall ResetToDefault_4702" );
		netconf_reset_fw ();
		WriteFile ( "/usr/bin/killall assoclistd" );

		CloseFile ();	/* Added by PeterChen */
		Executing ();	/* Added by PeterChen */

}

void start_wds(void)
{
	/* Bring up WDS interface */	
	sleep(3);
	system("/sbin/ifconfig wds0.2 up");
	system("/sbin/ifconfig wds0.3 up");
	system("/sbin/ifconfig wds0.4 up");
	system("/sbin/ifconfig wds0.5 up");
	system("/usr/sbin/brctl addif br0 wds0.2");
	system("/usr/sbin/brctl addif br0 wds0.3");
	system("/usr/sbin/brctl addif br0 wds0.4");
	system("/usr/sbin/brctl addif br0 wds0.5");
}

void main_loop_start_1(void)
{
		OpenFile ();	/* Added by PeterChen */
}

void main_loop_start_2(void)
{
	char	*wan_gateway = NULL;		// andrew add 20030529
	char	string[40];			// andrew add 20030529

	memset ( string , '\0' , 40 );
	
	start_httpd ();

	{	
		/*after filter nvram_set is no use*/
		GetWirelessMAC ();

		gemtek_restore_defaults_with_reboot();

		WriteFile ( "/sbin/assoclistd &" );
	}

	/* Fall through */
	/* Modify by Terence for Motorola */
	if ( !strcmp ( nvram_safe_get ( "WAN_ENABLE" ) , "1" ) ) {
		//printf("rc : ifconfig eth1 up...Do nothing\n");
		LED_wan_if_enable ( 1 );
		//system("ifconfig eth1 up");
	} else {
		//printf("rc : ifconfig eth1 down\n");
		LED_wan_if_enable ( 0 );
		//system("ifconfig eth1 down");			
		ifconfig ( nvram_safe_get ( "wan_ifname" ) , 0 , NULL , NULL );
	}			

	CloseFile ();	/* Added by PeterChen */
	Executing ();	/* Added by PeterChen */

	start_nas ("lan");	
	
	start_wds();	
	
	Peter_Added_001_H_Extra_SRCs_002	
}
